/***
 * Copyright (C) 1997,98 -- Miroslav Zagorac  <zaga@fly.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>

#include "lib_debug.h"
#include "lib_pcx.h"
#include "lib_string.h"


#ifdef WORDS_BIGENDIAN

#ifdef ANSI_FUNC

static void pcx_header_swap (struct PCX_HEADER *pcx_header)
#else
   
static void pcx_header_swap (pcx_header)
   struct PCX_HEADER *pcx_header;
#endif
{
   (void) swapw (&pcx_header->xmin);
   (void) swapw (&pcx_header->ymin);
   (void) swapw (&pcx_header->xmax);
   (void) swapw (&pcx_header->ymax);
   (void) swapw (&pcx_header->hres);
   (void) swapw (&pcx_header->vres);
   (void) swapw (&pcx_header->bytes_per_line);
   (void) swapw (&pcx_header->palette_type);
}

#endif /* WORDS_BIGENDIAN */


#ifdef ANSI_FUNC

struct IMAGE_DATA *pcx_image_load (char *filename)
#else

struct IMAGE_DATA *pcx_image_load (filename)
   char *filename;
#endif
/***
 * XXX  Funkcija radi dobro samo za 8-bitne slike.
 */
{
   struct PCX_HEADER  pcx_header;
   struct IMAGE_DATA *image = NULL;
   FILE              *pcx_fd = NULL;
   palette_t          buffer[SIZEOF_PALETTE];
   off_t              n, image_size;
   short              height, width;

   debug (("{%s(\"%s\")", __FUNCTION__, filename))

   if (! (pcx_fd = fopen (filename, "rb")))
   {
      image_error (filename, errno, &image, pcx_fd);

      debug_return ("} == 0x%08x", NULL)
   }

   if (fread (&pcx_header, SIZEOF_PCX_HEADER, 1, pcx_fd) != 1)
   {
      image_error (filename, errno, &image, pcx_fd);

      debug_return ("} == 0x%08x", NULL)
   }

#ifdef WORDS_BIGENDIAN
   pcx_header_swap (&pcx_header);
#endif

   width  = (pcx_header.xmax - pcx_header.xmin) + 1;
   height = (pcx_header.ymax - pcx_header.ymin) + 1;

   image_size = pcx_header.bytes_per_line * height * pcx_header.color_planes;

   debug (("----------------------------------------\n"))
   debug (("PCX image:  \"%s\"\n", filename))
   debug (("  signature      = %d\n", (int) pcx_header.signature))
   debug (("  version        = %d\n", (int) pcx_header.version))
   debug (("  encoding       = %d\n", (int) pcx_header.encoding))
   debug (("  bits_per_pixel = %d\n", (int) pcx_header.bits_per_pixel))
   debug (("  xmin           = %d\n", (int) pcx_header.xmin))
   debug (("  ymin           = %d\n", (int) pcx_header.ymin))
   debug (("  xmax           = %d\n", (int) pcx_header.xmax))
   debug (("  ymax           = %d\n", (int) pcx_header.ymax))
   debug (("  hres           = %d\n", (int) pcx_header.hres))
   debug (("  vres           = %d\n", (int) pcx_header.vres))
   debug (("  reserved       = %d\n", (int) pcx_header.reserved))
   debug (("  color_planes   = %d\n", (int) pcx_header.color_planes))
   debug (("  bytes_per_line = %d\n", (int) pcx_header.bytes_per_line))
   debug (("  palette_type   = %d\n", (int) pcx_header.palette_type))
   debug (("----------------------------------------\n"))

   if ((pcx_header.signature != PCX_MAGIC) || (pcx_header.version > PCX_LAST_VER))
   {
      image_error ("Not a PCX file!", 0, &image, pcx_fd);

      debug_return ("} == 0x%08x", NULL)
   }

   if (pcx_header.encoding != PCX_RLL)
   {
      image_error ("Not run-length encoded PCX file!", 0, &image, pcx_fd);

      debug_return ("} == 0x%08x", NULL)
   }

   if ((image = image_alloc (width, height)) == NULL)
   {
      image_error ("malloc(image)", errno, &image, pcx_fd);

      debug_return ("} == 0x%08x", NULL)
   }

   if (pcx_header.color_planes > PCX_MAXDEPTH)
   {
      image_error ("Cannot handle picture depth > 8!", 0, &image, pcx_fd);

      debug_return ("} == 0x%08x", NULL)
   }
   else if (pcx_header.color_planes < 8)
   {
      short flag = 0;

      (void) memcpy (image->palette, pcx_header.palette16, 16 * 3);

      for (n = 0; n < SIZEOF_PALETTE; n++)
         flag += image->palette[n];

      /***
       * if the color palette contained all zeros, set to black and white as default colors.
       */
      if (! flag)
      {
         image->palette[0] = image->palette[1] = image->palette[2] = 0x00;
         image->palette[3] = image->palette[4] = image->palette[5] = 0xff;
      }
   }

   /***
    * Load RLE data into image memory...
    */
   for (n = 0; n < image_size; )
   {
      int value = getc (pcx_fd);

      if (value < 0)
      {
         image_error (filename, errno, &image, pcx_fd);

         debug_return ("} == 0x%08x", NULL)
      }

      /***
       * Goes like this: Read a byte.  If the two high bits are set,
       * then the low 6 bits contain a repeat count, and the byte to
       * repeat is the next byte in the file.  If the two high bits are
       * not set, then this is the byte.
       */
      if ((value & 0xc0) == 0xc0)
      {
         unchar counter = value & 0x3f;

         if ((value = getc (pcx_fd)) < 0)
         {
            image_error (filename, errno, &image, pcx_fd);

            debug_return ("} == 0x%08x", NULL)
         }

         for ( ; counter && (n < image_size); counter--)
            image->image[n++] = (unchar) value;

         if (counter)
         {
            image_error ("Repeat count spans end of image!", 0, &image, pcx_fd);

            debug_return ("} == 0x%08x", NULL)
         }
      }
      else
         image->image[n++] = (unchar) value;
   }

   if (getc (pcx_fd) == PCX_256COLORS_MAGIC)
   {
      if (fread (buffer, SIZEOF_PALETTE, 1, pcx_fd) != 1)
      {
         image_error ("Image have no palette!", 0, &image, pcx_fd);

         debug_return ("} == 0x%08x", NULL)
      }

      (void) memcpy (image->palette, buffer, SIZEOF_PALETTE);
   }

   (void) fclose (pcx_fd);

   debug_return ("} == 0x%08x", image)
}


#ifdef ANSI_FUNC

image_t *pcx_get_image_part (struct IMAGE_DATA *image, short x_offset, short y_offset, short width, short height)
#else

image_t *pcx_get_image_part (image, x_offset, y_offset, width, height)
   struct IMAGE_DATA *image;
   short x_offset;
   short y_offset;
   short width;
   short height;
#endif
/***
 * XXX  Ovo radi dobro samo za 8-bitne slike.
 */
{
   static image_t *image_buffer = NULL;

   debug (("{%s(0x%08lx, %d, %d, %d, %d)", __FUNCTION__, (long) image, x_offset, y_offset, width, height))

   if (image_buffer)
   {
      free (image_buffer);
      image_buffer = NULL;
   }

   if (image && image->image)
   {
      image_t *image_ptr = image_buffer, *pcx_ptr = image->image + y_offset * image->width + x_offset;

      if (! (image_buffer = (image_t *) malloc (((long) width) * height)))
         debug_return ("} == 0x%08x", NULL)

      while (height--)
      {
         (void) memcpy (image_ptr, pcx_ptr, width);

         image_ptr += width;
         pcx_ptr   += image->width;
      }
   }

   debug_return ("} == 0x%08x", image_buffer)
}


#ifdef ANSI_FUNC

static unchar get_count (image_t *image, long *n, long image_size)
#else

static unchar get_count (image, n, image_size)
   image_t *image;
   long *n;
   long image_size;
#endif
{
   unchar count;

   debug (("{%s(0x%08lx, 0x%08lx, %ld)", __FUNCTION__, (long) image, (long) n, image_size))

   /***
    * Da, da, nije optimizirano bas, no to me nije ni briga.
    * Ako se zeli razumjeti sto se ovdje radi, moze se pogledati funkcija pcx_image_load().
    */
   if (((image[*n] & 0xc0) != 0xc0) && (image[*n + 1] != image[*n]))
   {
      count = image[*n];
      (*n)++;
   }
   else
   {
      for (count = 0x01; (*n + count) < image_size; count++)
         if (count == 0x3f)
            break;
         else if (image[*n] != image[*n + count])
            break;

      *n    += count;
      count |= 0xc0;
   }

   debug (("count = 0x%02x, n = %ld (of %ld)\n", (int) count, *n, image_size))

   debug_return ("} == %d", count)
}


#ifdef ANSI_FUNC

short pcx_image_save (struct IMAGE_DATA *image, char *filename)
#else

short pcx_image_save (image, filename)
   struct IMAGE_DATA *image;
   char *filename;
#endif
{
   struct PCX_HEADER  pcx_header;
   FILE              *pcx_fd = NULL;
   long               image_size, n;

   debug (("{%s(0x%08lx, \"%s\")", __FUNCTION__, (long) image, filename))

   if (!image || !image->image)
      debug_return ("} == %d", FUNC_RET_ERROR)

   if ((pcx_fd = fopen (filename, "w+b")) == NULL)
   {
      image_error (filename, errno, NULL, pcx_fd);

      debug_return ("} == %d", FUNC_RET_ERROR)
   }

   (void) memset (&pcx_header, 0, sizeof (struct PCX_HEADER));

   pcx_header.signature      = PCX_MAGIC;
   pcx_header.version        = PCX_LAST_VER;
   pcx_header.encoding       = PCX_RLL;
   pcx_header.bits_per_pixel = PCX_MAXDEPTH;
   pcx_header.xmin           = 0;
   pcx_header.ymin           = 0;
   pcx_header.xmax           = image->width - 1;
   pcx_header.ymax           = image->height - 1;
   pcx_header.hres           = 96;    /* XXX  Eh, da... */
   pcx_header.vres           = 96;
/* pcx_header.palette16[16][3] */
   pcx_header.reserved       = 0;
   pcx_header.color_planes   = 1;
   pcx_header.bytes_per_line = image->width;
   pcx_header.palette_type   = 1;
/* pcx_header.filler[58] */

   image_size = image->width * image->height;

#ifdef WORDS_BIGENDIAN
   pcx_header_swap (&pcx_header);
#endif

   if (fwrite (&pcx_header, sizeof (struct PCX_HEADER), 1, pcx_fd) != 1)
   {
      image_error (filename, errno, NULL, pcx_fd);

      debug_return ("} == %d", FUNC_RET_ERROR)
   }

   for (n = 0; n < image_size; )
   {
      unchar count = get_count (image->image, &n, image_size);

      (void) fputc (count, pcx_fd);

      if ((count & 0xc0) == 0xc0)
         (void) fputc (image->image[n - 1], pcx_fd);

      if (n > image_size)
      {
         image_error ("Internal error in function get_count()!", -1, NULL, pcx_fd);

         debug_return ("} == %d", FUNC_RET_ERROR)
      }
   }

   /***
    * Eh, ovdje mi se neda kontrolirati da li je jedan pisljivi byte snimljen, kad se
    * to radi i prije i poslije toga... >B)
    */
   (void) fputc (PCX_256COLORS_MAGIC, pcx_fd);

   if (fwrite (image->palette, 1, SIZEOF_PALETTE, pcx_fd) != SIZEOF_PALETTE)
   {
      image_error (filename, errno, NULL, pcx_fd);

      debug_return ("} == %d", FUNC_RET_ERROR)
   }

   (void) fclose (pcx_fd);

   debug_return ("} == %d", FUNC_RET_OK)
}
