/***
 * Copyright (C) 1996-98. by Miroslav Zagorac <zaga@tirnanog.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdarg.h>

#include "../lib/lib_debug.h"

#include "graphics.h"
#include "main.h"


#ifdef USE_X11

struct
{
   short x0;
   short x1;
   short y0;
   short y1;
} clip = { 0, WINDOW_WIDTH, 0, WINDOW_HEIGHT };


#ifdef ANSI_FUNC

static inline short clip_check (short *x, short *y, short *width, short *height, unchar **image)
#else

static inline short clip_check (x, y, width, height, image)
   short *x;
   short *y;
   short *width;
   short *height;
   unchar **image;
#endif
{
   if (*y < clip.y0)
   {
      if (image && *image)
         *image  += (clip.y0 - *y) * *width;

      *height -= clip.y0 - *y;
      *y       = clip.y0;
   }
   else if ((*y + *height) > clip.y1)
      *height = clip.y1 - *y;

   if (*x < clip.x0)
   {
      if (image && *image)
         *image += clip.x0 - *x;

      *width -= clip.x0 - *x;
      *x      = clip.x0;
   }
   else if ((*x + *width) > clip.x1)
      *width = clip.x1 - *x;

   debug (("x,y = %d,%d    width,height = %d,%d\n", *x, *y, *width, *height))

   return ((*width > 0) && (*height > 0));
}

#endif /* USE_X11 */


#ifdef ANSI_FUNC

void graph_rectangle (short x, short y, short width, short height, short color_index)
#else

void graph_rectangle (x, y, width, height, color_index)
   short x;
   short y;
   short width;
   short height;
   short color_index;
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      svga_draw_rectangle (x, y, width, height, color_index);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
      X11_draw_rectangle (&main_window, x, y, width, height, color_index);
#endif
}


#ifdef ANSI_FUNC

void graph_putbox (short x, short y, short width, short height, unchar *image)
#else

void graph_putbox (x, y, width, height, image)
   short x;
   short y;
   short width;
   short height;
   unchar *image;
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      gl_putbox (x, y, width, height, image);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
   {
      short image_width = width;

      debug (("{%s(%d, %d, %d, %d, 0x%08lx)", __FUNCTION__, x, y, width, height, (long) image))

      if (! clip_check (&x, &y, &width, &height, &image))
         debug_noreturn

      if (x11_bpp == 1)
      {
         register unchar *screen_ptr = (unchar *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         register unchar *image_ptr = image;
         short            i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width, image_ptr += image_width - width)
            for (i = 0; i < width; i++)
               *(screen_ptr++) = *(image_ptr++);
      }
      else if (x11_bpp == 2)
      {
         register ushort *screen_ptr = (ushort *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         register unchar *image_ptr = image;
         short            i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width, image_ptr += image_width - width)
            for (i = 0; i < width; i++)
               *(screen_ptr++) = (ushort) Pixel_color[*(image_ptr++)];
      }
      else if ((x11_bpp == 3) || (x11_bpp == 4))
      {
         register ulong  *screen_ptr = (ulong *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         register unchar *image_ptr = image;
         short            i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width, image_ptr += image_width - width)
            for (i = 0; i < width; i++)
               *(screen_ptr++) = (ulong) Pixel_color[*(image_ptr++)];
      }

      debug_endfunc
   }
#endif
}


#ifdef ANSI_FUNC

void graph_putboxmask (short x, short y, short width, short height, unchar *image)
#else

void graph_putboxmask (x, y, width, height, image)
   short x;
   short y;
   short width;
   short height;
   unchar *image;
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      gl_putboxmask (x, y, width, height, image);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
   {
      debug (("{%s(%d, %d, %d, %d, 0x%08lx)", __FUNCTION__, x, y, width, height, (long) image))

      /***
       * Ovdje je funkcija slicna graph_putbox(), uz dodatak ispitivanja da li je pixel
       * koji se prikazuje proziran ili ne.
       */
      if (x11_bpp == 1)
      {
         register unchar *screen_ptr = (unchar *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         register unchar *image_ptr = image;
         short            i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width)
            for (i = 0; i < width; i++, screen_ptr++)
               if (*(image_ptr++))
                  *screen_ptr = *image_ptr;
      }
      else if (x11_bpp == 2)
      {
         register ushort *screen_ptr = (ushort *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         register unchar *image_ptr = image;
         short            i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width)
            for (i = 0; i < width; i++, screen_ptr++)
               if (*(image_ptr++))
                  *screen_ptr = (ushort) Pixel_color[*image_ptr];
      }
      else if ((x11_bpp == 3) || (x11_bpp == 4))
      {
         register ulong  *screen_ptr = (ulong *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         register unchar *image_ptr = image;
         short            i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width)
            for (i = 0; i < width; i++, screen_ptr++)
               if (*(image_ptr++))
                  *screen_ptr = (ulong) Pixel_color[*image_ptr];
      }

      debug_endfunc
   }
#endif
}


#ifdef ANSI_FUNC

void graph_fillbox (short x, short y, short width, short height, short color_index)
#else

void graph_fillbox (x, y, width, height, color_index)
   short x;
   short y;
   short width;
   short height;
   short color_index;
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      gl_fillbox (x, y, width, height, color_index);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
   {
      debug (("{%s(%d, %d, %d, %d, %d)", __FUNCTION__, x, y, width, height, color_index))

      if (! clip_check (&x, &y, &width, &height, NULL))
         debug_noreturn

      if (x11_bpp == 1)
      {
         register unchar *screen_ptr = (unchar *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         short            i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width)
            for (i = 0; i < width; i++)
               *(screen_ptr++) = (unchar) color_index;
      }
      else if (x11_bpp == 2)
      {
         register ushort *screen_ptr = (ushort *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         register ushort  pixel = (ushort) Pixel_color[color_index], i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width)
            for (i = 0; i < width; i++)
               *(screen_ptr++) = pixel;
      }
      else if ((x11_bpp == 3) || (x11_bpp == 4))
      {
         register ulong *screen_ptr = (ulong *) (main_window.shared_memory + (x + y * WINDOW_WIDTH) * x11_bpp);
         register ulong  pixel = (ulong) Pixel_color[color_index], i;

         for ( ; height--; screen_ptr += WINDOW_WIDTH - width)
            for (i = 0; i < width; i++)
               *(screen_ptr++) = pixel;
      }

      debug_endfunc
   }
#endif
}


#ifdef ANSI_FUNC

void graph_scalebox (short width, short height, unchar *image, short width_1, short height_1, unchar *image_1)
#else

void graph_scalebox (width, height, image, width_1, height_1, image_1)
   short width;
   short height;
   unchar *image;
   short width_1;
   short height_1;
   unchar *image_1;
#endif
{
   if (image && image_1)
   {
#ifdef USE_SVGALIB
      if (game_data.graphics_type == GRAPHICS_SVGALIB)
         gl_scalebox (width, height, image, width_1, height_1, image_1);
#endif

#ifdef USE_X11
      if (game_data.graphics_type == GRAPHICS_X11)
      {
         /***
          * This function was adapted from svgalib, by Harm Hanemaayer.  Thanks!
          * Work only with 8bpp images.
          */
         unchar *dp1 = image, *dp2 = image_1;
         int     xfactor, yfactor, y, sy = 0;

         if (!width_1 || !height_1)
            return;

         xfactor = width * 65536L / width_1;	/* scaled by 65536 */
         yfactor = height * 65536L / height_1;	/* scaled by 65536 */

         for (y = 0; y < height_1; )
         {
            unchar *dp2old = dp2;
            int     sx = 0, x;

            for (x = 0; x < width_1 - 8; x += 8)
            {
               *(dp2 + x) = *(dp1 + (sx >> 16));
               sx += xfactor;

               *(dp2 + x + 1) = *(dp1 + (sx >> 16));
               sx += xfactor;

               *(dp2 + x + 2) = *(dp1 + (sx >> 16));
               sx += xfactor;

               *(dp2 + x + 3) = *(dp1 + (sx >> 16));
               sx += xfactor;

               *(dp2 + x + 4) = *(dp1 + (sx >> 16));
               sx += xfactor;

               *(dp2 + x + 5) = *(dp1 + (sx >> 16));
               sx += xfactor;

               *(dp2 + x + 6) = *(dp1 + (sx >> 16));
               sx += xfactor;

               *(dp2 + x + 7) = *(dp1 + (sx >> 16));
               sx += xfactor;
            }

            while (x < width_1)
            {
               *(dp2 + x++) = *(dp1 + (sx >> 16));
               sx += xfactor;
            }

            dp2 += width_1;

            for (y++; y < height_1; y++)
            {
               int l, syint = sy >> 16;

               if (((sy += yfactor) >> 16) != syint)
                  break;

               /***
                * Copy identical lines.
                */
               l = dp2 - dp2old;

               memcpy (dp2, dp2old, l);

               dp2old = dp2;
               dp2 += l;
            }

            dp1 = image + (sy >> 16) * width;
         }
      }
#endif
   }
}


#ifdef ANSI_FUNC

void graph_clearscreen (short color_index)
#else

void graph_clearscreen (color_index)
   short color_index;
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      gl_clearscreen (color_index);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
      (void) memset (main_window.shared_memory, Pixel_color[color_index], main_window.shared_len);
#endif
}


#ifdef ANSI_FUNC

void graph_setpalettecolor (short color_index, short red, short green, short blue)
#else

void graph_setpalettecolor (color_index, red, green, blue)
   short color_index;
   short red;
   short green;
   short blue;
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      gl_setpalettecolor (color_index, red, green, blue);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
      X11_set_color (red, green, blue, color_index);
#endif
}


#ifdef ANSI_FUNC

void graph_setclippingwindow (short x_0, short y_0, short x_1, short y_1)
#else

void graph_setclippingwindow (x_0, y_0, x_1, y_1)
   short x_0;
   short y_0;
   short x_1;
   short y_1;
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      gl_setclippingwindow (x_0, y_0, x_1, y_1);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
   {
      clip.x0 = x_0;
      clip.x1 = x_1 + 1;
      clip.y0 = y_0;
      clip.y1 = y_1 + 1;
   }
#endif
}


#ifdef ANSI_FUNC

void graph_palette_fade (short direction)
#else

void graph_palette_fade (direction)
   short direction;
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      svga_palette_fade (direction);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
      X11_colormap_fade (direction);
#endif
}


#ifdef ANSI_FUNC

void graph_copyscreen (void)
#else

void graph_copyscreen ()
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      svga_copyscreen (TEST_OPTION (OPT_WAITRETRACE) ? 0 : 1);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
      X11_copyscreen (&main_window);
#endif
}


#ifdef ANSI_FUNC

short graph_save_snapshot (void)
#else

short graph_save_snapshot ()
#endif
{
#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      return (svga_save_snapshot ());
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
      return (X11_save_snapshot (main_window.window));
#endif

   return (FUNC_RET_ERROR);
}
    
    
#ifdef ANSI_FUNC

void graph_printf (short x, short y, short color, struct FONT_DATA *font, char *format, ...)
#else

void graph_printf (x, y, color, font, format)
   short x;
   short y;
   short color;
   struct FONT_DATA *font;
   char *format; /* XXX varargs */
#endif
{   
   va_list ap;
   char    buffer[BUFSIZ];

   debug (("{%s(%d, %d, %d, 0x%08lx...)", __FUNCTION__, x, y, color, (long) font));
 
   va_start (ap, format);
   (void) vsprintf (buffer, format, ap);
   va_end (ap);

#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      svga_printf (x, y, color, font, buffer);
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
   {
      short i, font_charsize = font->width * font->height;

      for (i = 0; buffer[i]; i++)
      {
         font_char_color (font, color, buffer[i]);
         graph_putbox (x + i * font->width, y, font->width, font->height, font->bitmap + buffer[i] * font_charsize);
      }
   }
#endif

   debug_endfunc
}   
